package pp.s1184725.boppi;

import java.util.EmptyStackException;
import java.util.logging.Logger;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.ParseTree;

import pp.s1184725.boppi.BasicParser.*;

/**
 * This class performs type checking and variable assignment on a bare parse
 * tree.
 *
 */
public class BoppiBasicChecker extends BasicBaseVisitor<SimpleType> {
	private CachingSymbolTable<SimpleType> symbols;
	private Annotations an;
	private Logger log;

	/**
	 * Creates a new checker with a logger and an object to link types and
	 * variables to parse tree nodes.
	 * 
	 * @param logger
	 *            the logger object to write warnings and errors to
	 * @param annotations
	 *            the annotations object
	 */
	public BoppiBasicChecker(Logger logger, Annotations annotations) {
		symbols = new CachingSymbolTable<>();
		an = annotations;
		log = logger;
	}

	@Override
	public SimpleType visit(ParseTree tree) {
		SimpleType type = super.visit(tree);
		an.types.put(tree, type);
		return type;
	}

	@Override
	public SimpleType visitAssign(AssignContext ctx) {
		SimpleType expr = visit(ctx.singleExpr());
		SimpleType var = visit(ctx.variable());
		checkConstraint(expr, var, ctx);
		return expr;
	}

	@Override
	public SimpleType visitBool(BoolContext ctx) {
		return SimpleType.BOOL;
	}

	@Override
	public SimpleType visitBlock(BlockContext ctx) {
		return symbols.withScope(() -> visit(ctx.expr()));
	}

	@Override
	public SimpleType visitChar(CharContext ctx) {
		return SimpleType.CHAR;
	}

	@Override
	public SimpleType visitExpr(ExprContext ctx) {
		if (ctx.singleExpr(ctx.singleExpr().size() - 1) instanceof DeclareContext)
			log.severe(getError(ctx, "Compound expression ends with declaration."));

		return ctx.singleExpr().stream().map(this::visit).reduce((__, snd) -> snd).get();
	}

	@Override
	public SimpleType visitDeclare(DeclareContext ctx) {
		try {
			Variable<SimpleType> var = symbols.put(ctx.IDENTIFIER().getText(), visit(ctx.type()));
			an.variables.put(ctx, var);
		} catch (EmptyStackException e) {
			log.severe(getError(ctx, "Declaring variable outside a program."));
		} catch (Exception e) {
			log.severe(getError(ctx, e.getMessage()));
		}
		return SimpleType.VOID;
	}

	@Override
	public SimpleType visitIf(IfContext ctx) {
		return symbols.withScope(() -> {
			visit(ctx.cond);
			if (ctx.onFalse != null) {
				SimpleType trueType = symbols.withScope(() -> visit(ctx.onTrue));
				SimpleType falseType = symbols.withScope(() -> visit(ctx.onFalse));

				return (trueType.equals(falseType)) ? trueType : SimpleType.VOID;
			} else {
				symbols.withScope(() -> visit(ctx.onTrue));
				return SimpleType.VOID;
			}
		});
	}

	@Override
	public SimpleType visitInfix1(Infix1Context ctx) {
		checkConstraint(visit(ctx.lhs), SimpleType.INT, ctx.lhs);
		checkConstraint(visit(ctx.rhs), SimpleType.INT, ctx.rhs);

		return SimpleType.INT;
	}

	@Override
	public SimpleType visitInfix2(Infix2Context ctx) {
		checkConstraint(visit(ctx.lhs), SimpleType.INT, ctx.lhs);
		checkConstraint(visit(ctx.rhs), SimpleType.INT, ctx.rhs);

		return SimpleType.INT;
	}

	@Override
	public SimpleType visitInfix3(Infix3Context ctx) {
		SimpleType lht = visit(ctx.lhs);
		SimpleType rht = visit(ctx.rhs);
		checkConstraint(lht, rht, ctx);

		switch (ctx.op.getType()) {
		case BasicLexer.LT:
		case BasicLexer.LEQ:
		case BasicLexer.GTE:
		case BasicLexer.GT:
			checkConstraint(lht, SimpleType.INT, ctx);
		}

		return SimpleType.BOOL;
	}

	@Override
	public SimpleType visitInfix4(Infix4Context ctx) {
		checkConstraint(visit(ctx.lhs), SimpleType.BOOL, ctx.lhs);
		checkConstraint(visit(ctx.rhs), SimpleType.BOOL, ctx.rhs);

		return SimpleType.BOOL;
	}

	@Override
	public SimpleType visitInfix5(Infix5Context ctx) {
		checkConstraint(visit(ctx.lhs), SimpleType.BOOL, ctx.lhs);
		checkConstraint(visit(ctx.rhs), SimpleType.BOOL, ctx.rhs);

		return SimpleType.BOOL;
	}

	@Override
	public SimpleType visitNumber(NumberContext ctx) {
		return SimpleType.INT;
	}

	@Override
	public SimpleType visitParens(ParensContext ctx) {
		return visit(ctx.expr());
	}

	@Override
	public SimpleType visitPrefix1(Prefix1Context ctx) {
		SimpleType type = visit(ctx.singleExpr());

		switch (ctx.op.getType()) {
		case BasicLexer.NOT:
			checkConstraint(type, SimpleType.BOOL, ctx.singleExpr());
			break;

		case BasicLexer.PLUS:
		case BasicLexer.MINUS:
			checkConstraint(type, SimpleType.INT, ctx.singleExpr());
			break;
		}

		return type;
	}

	@Override
	public SimpleType visitProgram(ProgramContext ctx) {
		symbols.withScope(() -> super.visitProgram(ctx));
		return null;
	}

	@Override
	public SimpleType visitRead(ReadContext ctx) {
		if (ctx.variable().size() == 1) {
			return visit(ctx.variable(0));
		} else {
			ctx.variable().forEach(this::visit);
			return SimpleType.VOID;
		}
	}

	@Override
	public SimpleType visitType(TypeContext ctx) {
		if (ctx.variable() != null)
			return visit(ctx.variable());
		else
			return SimpleType.fromToken(ctx.staticType.getType());
	}

	@Override
	public SimpleType visitVar(VarContext ctx) {
		return visit(ctx.variable());
	}

	@Override
	public SimpleType visitVariable(VariableContext ctx) {
		try {
			Variable<SimpleType> var = symbols.get(ctx.getText());
			an.variables.put(ctx, var);
			return var.getType();
		} catch (EmptyStackException e) {
			log.severe(getError(ctx, "Using variable outside a program."));
		} catch (Exception e) {
			log.severe(getError(ctx, e.getMessage()));
		}

		return SimpleType.VOID;
	}

	@Override
	public SimpleType visitWhile(WhileContext ctx) {
		return symbols.withScope(() -> {
			checkConstraint(visit(ctx.cond), SimpleType.BOOL, ctx.cond);
			symbols.withScope(() -> visit(ctx.onTrue));
			return SimpleType.VOID;
		});
	}

	@Override
	public SimpleType visitWrite(WriteContext ctx) {
		if (ctx.expr().size() == 1) {
			SimpleType type = visit(ctx.expr(0));
			if (SimpleType.VOID.equals(type))
				log.severe(getError(ctx, "Cannot print argument of type %s.", type));

			return type;
		} else {
			ctx.expr().stream().map(this::visit).forEach((type) -> {
				if (SimpleType.VOID.equals(type))
					log.severe(getError(ctx, "Cannot print argument of type %s.", type));
			});
			return SimpleType.VOID;
		}
	}

	/**
	 * Checks whether two types are compatible and log an error if they do not.
	 * 
	 * @param type1
	 *            the left hand or actual type
	 * @param type2
	 *            the right hand or desired type
	 * @param node
	 *            the context used for logging
	 */
	private void checkConstraint(SimpleType type1, SimpleType type2, ParserRuleContext node) {
		if (!type2.equals(type1))
			log.severe(getError(node, "Could not match type %s with type %s.", type1.toString(), type2.toString()));
	}

	/**
	 * Returns an error message for a given parse tree node.
	 * 
	 * @param ctx
	 *            the parse tree node at which the error occurred
	 * @param message
	 *            the error message
	 * @param args
	 *            arguments for the message, see {@link String#format}
	 */
	private String getError(ParserRuleContext node, String message, Object... args) {
		int line = node.getStart().getLine();
		int column = node.getStart().getCharPositionInLine();
		return String.format("Line %d:%d - %s", line, column, String.format(message, args));
	}
}
